/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

#include <cstring>
#include <algorithm>
using std::swap;

void Hilbert3DCurve::analyseGates()
{
  properties.reject(VertexGated);
  properties.reject(VertexEdgeGated);
  properties.reject(VertexFacetGated);
  properties.reject(EdgeGated);
  properties.reject(FacetGated);
  properties.reject(EdgeCrossing);
  properties.reject(FacetCrossing);
  properties.reject(CubeCrossing);
  
  switch (theName[5])
  {
    case 'c':
      switch (theName[6])
      {
        case 'c':                     properties.confirm(VertexGated); 
                                      if (collinear(basePattern[0], basePattern[7])) properties.confirm(EdgeCrossing);
                                      else                                           properties.confirm(FacetCrossing);
                                      break;
        case 'r': case 't': case 'v': properties.confirm(VertexEdgeGated);  properties.confirm(FacetCrossing); break;
        case 'f': case 'g': case 's': properties.confirm(VertexFacetGated); properties.confirm(CubeCrossing); break;
      }
      break;
    case 'r': case 't': case 'v':
      switch (theName[6])
      {
        case 'c':                     properties.confirm(VertexEdgeGated);  properties.confirm(FacetCrossing); break;
        case 'r': case 't': case 'v': properties.confirm(EdgeGated);        properties.confirm(CubeCrossing); break;
        case 'f': case 'g': case 's': properties.confirm(Invalid); break;
      }
      break;
    case 'f': case 'g': case 's':
      switch (theName[6])
      {
        case 'c':                     properties.confirm(VertexFacetGated); properties.confirm(CubeCrossing); break;
        case 'r': case 't': case 'v': properties.confirm(Invalid); break;
        case 'f': case 'g': case 's': properties.confirm(FacetGated);       properties.confirm(CubeCrossing); break;
      }
      break;
  }    
}


void Hilbert3DCurve::analyseTraversal(const OutputLevel testLevel)
{
  // determines the following properties:
  // FaceContinuous, Hyperorthogonal, FacetHarmonious, DiagonalHarmonious

    // if none, only the simplest properties are tested for, there will be no report on 
    // diagonal-harmony, unless the filter requires this;
    // facet-harmony is hardcoded and not verified;
    // if concise (default), diagonal-harmony are determined; facet-harmony is verified;

  properties.reject(FacetHarmonious);
  if (theName == "Ca00.cc.44.II.ZZ") properties.confirm(FacetHarmonious);

  properties.reject(FaceContinuous);
  properties.reject(Hyperorthogonal);
  if (theName.substr(0,7) == "Ca00.cc" || theName.substr(0,7) == "Si00.cc" || theName == "Ca00.gs") 
  {
    properties.confirm(FaceContinuous);
    if (theName == "Ca00.gs" || theName == "Ca00.cc.44.hh.db") properties.confirm(Hyperorthogonal);
  }
  
  if (testLevel == None) return;
  
  properties.reject(DiagonalHarmonious);
  
  vector<OPt> v;
  v.clear();
  
  struct Sequence
  {
    vector<OPt> v;
    vector<OPt>::iterator f;
    Sequence() {};
    Sequence& add(const OPt& p) { v.push_back(p); return *this; }
    void start() { f = v.begin(); }
    void see(const OPt& p) { if (f != v.end() && *f == p) ++f; }
    bool inOrder() { return (f == v.end()); }
  };
  
  Sequence diagonal[8];
  diagonal[0].add(OPt(-3,-3,-3)).add(OPt(-1,-1,-1)).add(OPt(1,1,1)).add(OPt(3,3,3)).start();
  diagonal[1].add(OPt(-3,-3,3)).add(OPt(-1,-1,1)).add(OPt(1,1,-1)).add(OPt(3,3,-3)).start();
  diagonal[2].add(OPt(-3,3,-3)).add(OPt(-1,1,-1)).add(OPt(1,-1,1)).add(OPt(3,-3,3)).start();
  diagonal[3].add(OPt(-3,3,3)).add(OPt(-1,1,1)).add(OPt(1,-1,-1)).add(OPt(3,-3,-3)).start();
  diagonal[4].add(OPt(3,-3,-3)).add(OPt(1,-1,-1)).add(OPt(-1,1,1)).add(OPt(-3,3,3)).start();
  diagonal[5].add(OPt(3,-3,3)).add(OPt(1,-1,1)).add(OPt(-1,1,-1)).add(OPt(-3,3,-3)).start();
  diagonal[6].add(OPt(3,3,-3)).add(OPt(1,1,-1)).add(OPt(-1,-1,1)).add(OPt(-3,-3,3)).start();
  diagonal[7].add(OPt(3,3,3)).add(OPt(1,1,1)).add(OPt(-1,-1,-1)).add(OPt(-3,-3,-3)).start();
  
  vector<OPt> hilbert2D;
  hilbert2D.push_back(OPt(-3,-3,-3)); hilbert2D.push_back(OPt(-3,-3,-1)); hilbert2D.push_back(OPt(-3,-1,-1)); hilbert2D.push_back(OPt(-3,-1,-3));
  hilbert2D.push_back(OPt(-3, 1,-3)); hilbert2D.push_back(OPt(-3, 3,-3)); hilbert2D.push_back(OPt(-3, 3,-1)); hilbert2D.push_back(OPt(-3, 1,-1));
  hilbert2D.push_back(OPt(-3, 1, 1)); hilbert2D.push_back(OPt(-3, 3, 1)); hilbert2D.push_back(OPt(-3, 3, 3)); hilbert2D.push_back(OPt(-3, 1, 3));
  hilbert2D.push_back(OPt(-3,-1, 3)); hilbert2D.push_back(OPt(-3,-1, 1)); hilbert2D.push_back(OPt(-3,-3, 1)); hilbert2D.push_back(OPt(-3,-3, 3));
  Sequence hilbert[48];
  for (int tid = 0; tid < 48; ++tid)
  {
    GeometricTransformation t(tid);
    for (int i = 0; i < 16; ++i) hilbert[tid].add(t(hilbert2D[i]));
    hilbert[tid].start();
  }

  struct InternalFace
  {
    stack<OPt> s;
    OPt edgePoint;
    InternalFace(const OPt& p): edgePoint(p) {};
    void see(const OPt& p) 
    { 
      bool onFace = true;
      bool onStack = !s.empty();
      for (int i = 0; onFace && i < d; ++i)
        switch(edgePoint[i])
        {
          case -3: case 3: 
            onFace = ((p[i] ^ edgePoint[i]) >= 0);      // i.e. having same sign
            onStack = (onStack && s.top()[i] == p[i]);
            break;
          case -1: case 1: 
            onFace = (p[i] == -1 || p[i] == 1); 
            break;
        }
      if (!onFace) return;
      if (onStack) s.pop(); else s.push(p);
    }
    bool inOrder() { return (s.empty()); }
  };
  InternalFace internalFace[12] = {
    InternalFace(OPt(-3,-3,-1)), InternalFace(OPt(-3,3,-1)), InternalFace(OPt(3,-3,-1)), InternalFace(OPt(3,3,-1)),
    InternalFace(OPt(-3,-1,-3)), InternalFace(OPt(-3,-1,3)), InternalFace(OPt(3,-1,-3)), InternalFace(OPt(3,-1,3)),
    InternalFace(OPt(-1,-3,-3)), InternalFace(OPt(-1,-3,3)), InternalFace(OPt(-1,3,-3)), InternalFace(OPt(-1,3,3))
  };
  
  generateTraversal(v, 2); 
  for (int i = 0; i < (1 << 6); ++i)
  {
    for (int j = 0; j < 8; ++j) diagonal[j].see(v[i]);
    for (int j = 0; j < 48; ++j) hilbert[j].see(v[i]);
    for (int j = 0; j < 12; ++j) internalFace[j].see(v[i]);
  }  
    
  int harmoniousDiagonals = 0;
  for (int j = 0; j < 8; ++j) if (diagonal[j].inOrder()) ++harmoniousDiagonals;
  if (harmoniousDiagonals >= 4) properties.confirm(DiagonalHarmonious);
  
  int harmoniousFacets = 0;
  for (int j = 0; j < 48; ++j) if (hilbert[j].inOrder()) ++harmoniousFacets;
  if (harmoniousFacets >= 5)
  {
    if (!properties.is(FacetHarmonious) && theName != "Ca00.cc.TT.II.77")
    {
      cerr << theName << " passes facet-harmony test on A2, but should fail!" << endl;
      properties.confirm(Invalid);
    }
  }
  else
  {
    if (properties.is(FacetHarmonious) || theName == "Ca00.cc.TT.II.77")
    {
      cerr << theName << " fails facet-harmony test on A2, but should pass!" << endl;
      properties.confirm(Invalid);
    }
  }
  
  bool palindromic = true;
  for (int j = 0; palindromic && j < 12; ++j) palindromic = internalFace[j].inOrder();
  if (theName == "Ca00.cc.TT.II.77" || theName == "Ca00.cc.TT.II.99")
  {
    if (!palindromic) 
    { 
      cerr << theName << " fails palindromicness test on A2, but should pass!" << endl;
      properties.confirm(Invalid);
    }
  }
  else if (palindromic) 
  {
    cerr << theName << " passes palindromicness test, but, according to previous tests, it should fail!" << endl;
    properties.confirm(Invalid);
  }

  v.clear();

  if (testLevel != Full) return;

  generateTraversal(v, 3);
  bool faceContinuous = (sqDistance(v[0], v[1]) == 4);
  bool hyperorthogonal = true;
  for (int i = 2; faceContinuous && i < (1 << 9); ++i)
  {
    faceContinuous = (sqDistance(v[i-1], v[i]) == 4);
    hyperorthogonal = hyperorthogonal && (sqDistance(v[i-2], v[i]) == 8);
  }
  if (faceContinuous) 
  {
    if (!properties.is(FaceContinuous))
    {
      cerr << theName << " passes face-continuity test on A3, but should fail!" << endl;
      properties.confirm(Invalid);
    }
    if (hyperorthogonal && !properties.is(Hyperorthogonal)) 
    {
      cerr << theName << " passes hyperorthogonality test on A3, but should fail!" << endl;
      properties.confirm(Invalid);
    }
  }
  else
  {
    if (properties.is(FaceContinuous))
    {
      cerr << theName << " fails face-continuity test on A3, but should pass!" << endl;
      properties.confirm(Invalid);
    }
  }
}


void Hilbert3DCurve::analyseTransformations(const OutputLevel testLevel)
{
  properties.reject(Centred);
  properties.reject(OrderPreserving);
  properties.reject(CoordinateShifting);
  properties.reject(Standing);
  
  if (transformationPattern[3].direction == 0)
  {
    if (theName[6] == 'c' && (GPt) basePattern[3] + transformationPattern[3].geometry((GPt) basePattern[7]) == GPt(0,0,0))
      properties.confirm(Centred);
  }
  else
  {
    if (theName[5] == 'c' && (GPt) basePattern[3] + transformationPattern[3].geometry((GPt) basePattern[0]) == GPt(0,0,0))
      properties.confirm(Centred);
  }

  if (is(Symmetric)) 
    properties.confirm(OrderPreserving);
  else
  {  
    bool orderPreserving = true;
    for (int i = 0; orderPreserving && i < 8; ++i) orderPreserving = (transformationPattern[i].direction == 0);
    if (orderPreserving) properties.confirm(OrderPreserving);
  }
  
  bool unsignedPermutations[6]; for (int k = 0; k < 6; ++k) unsignedPermutations[k] = false;
  for (int i = 0; i < 8; ++i) unsignedPermutations[(int) transformationPattern[i].geometry >> 3] = true;
  int nrShifts = (unsignedPermutations[3]?1:0) + (unsignedPermutations[4]?1:0);
  int nrSwaps  = (unsignedPermutations[1]?1:0) + (unsignedPermutations[2]?1:0) + (unsignedPermutations[5]?1:0);
  if (nrSwaps == 0) properties.confirm(CoordinateShifting);
  if (nrShifts == 0 && nrSwaps <= 1) properties.confirm(Standing);
  
  // if not required, do not calculate pattern-isotropy
  if (testLevel == None) return;
  
  properties.reject(PatternIsotropic);
   
  int64 transitions[48];

  Symmetries::iterator sigma;
  for (sigma = transformations.begin(); sigma != transformations.end() && sigma->symbol != theName[1]; ++sigma)
  if (sigma == transformations.end()) cerr << "SOMETHING IS WRONG" << endl;
  
  for (int tid = 0; tid < 48; ++tid)
  { 
    transitions[tid] = 0;
    GeometricTransformation t(tid);
    for (int i = 0; i < 8; ++i) 
    {
      transitions[tid] |= (int64) 1 << (int) (transformationPattern[i].geometry * t);
      if (is(Symmetric)) transitions[tid] |= (int64) 1 << (int) (transformationPattern[i].geometry * sigma->geometry * t);
    }
  }
  
  vector<int64> gamma;
  gamma.push_back(1);
  int a; int b = 0;
  do
  {
    gamma.push_back(0); ++b;
    for (int t = 0; t < 48; ++t) gamma[b] |= (gamma[b-1] >> t & 1) * transitions[t];
    a = b - 1;
    while (a >= 0 && gamma[a] != gamma[b]) --a;
  }
  while (a < 0);
  
  if (gamma[a] == (int64) 0xffffffffffff) 
    properties.confirm(PatternIsotropic);
}

Hilbert3DCurve::Hilbert3DCurve(Hilbert3DCurveSearch& search, const OutputLevel testLevel)
{
  while (search.findValidCurve(*this))
  {
    properties.reject(Invalid);
    
    Transformation embedding;
    if (search.filter.matchTraversal(*this, embedding))
    {
      for (int i = 0; i < 8; ++i) 
      {
        basePattern[i] = embedding.geometry(basePattern[i]);
        transformationPattern[i] = embedding * transformationPattern[i] * embedding.inverse();
      }
      if (embedding.direction != 0)
        for (int i = 0; i < 4; ++i)
        {
          swap(basePattern[i], basePattern[i^7]);
          swap(transformationPattern[i], transformationPattern[i^7]);
        }
      for (int i = 0; i < 8; ++i)
        invBasePattern[convert(basePattern[i])[0]][convert(basePattern[i])[1]][convert(basePattern[i])[2]] = i;
    }
    else
    {
      properties.confirm(Invalid);
      search.skip(); 
      continue;
    }
    
    properties.reject(Symmetric);
    properties.reject(Metasymmetric);
    properties.reject(WellFolded);
   
    if (theName.length() == 16 && 
      theName[1] <= 'o' &&
      theName[2] == theName[3] &&
      theName[5] == theName[6] &&
      theName[8] == theName[9] &&
      theName[11] == 'I' && theName[12] == 'I' &&
      theName[14] == theName[15]
    ) 
    {
      properties.confirm(Symmetric);
      if (
        (theName.substr(0,10) == "Ca00.cc.TT" && string("bC47").find(theName[14]) != string::npos) ||
        (theName.substr(0,10) == "Ca11.cc.TT" && string("IPJ3").find(theName[14]) != string::npos) ||
        (theName.substr(0,10) == "Cd00.cc.PP" && string("IPJ3").find(theName[14]) != string::npos) ||
        (theName.substr(0,10) == "Cd11.cc.PP" && string("bC47").find(theName[14]) != string::npos) ||
        (theName.substr(0,10) == "Se00.cc.TT" && string("bPJ7").find(theName[14]) != string::npos) ||
        (theName.substr(0,10) == "Se66.cc.TT" && string("IC43").find(theName[14]) != string::npos)
      ) properties.confirm(Metasymmetric);
    }
    
    if (theName.substr(0,4) == "Ca00") properties.confirm(WellFolded);

    analyseGates();
    
    if (testLevel == None && (search.filter.forbids(DiagonalHarmonious) || search.filter.requires(DiagonalHarmonious)))
      analyseTraversal(Concise);
    else
      analyseTraversal(testLevel);

    if (testLevel == None && (search.filter.forbids(PatternIsotropic) || search.filter.requires(PatternIsotropic)))
      analyseTransformations(Concise);
    else
      analyseTransformations(testLevel);

    for (int i = 0; i < (int) Invalid; ++i)
      if ((is((Property) i) && search.filter.forbids((Property) i)) ||
          (isNot((Property) i) && search.filter.requires((Property) i)))
        properties.confirm(Invalid);    
      
    if (is(Invalid)) search.skip(); else return;
  }
  properties.confirm(Invalid);
  return;
}


bool Hilbert3DCurve::is(Property p) const
{
  return properties.is(p);
}


bool Hilbert3DCurve::isNot(Property p) const
{
  return properties.isNot(p);
}


string Hilbert3DCurve::name() const
{
  if (is(Invalid)) return string("Not a curve");
  return theName;
}


string Hilbert3DCurve::shortName() const
{
  if (is(Invalid)) return string("Not a curve");
  if (is(Symmetric)) {
    string n(theName, 0, 6);
    n += theName[8]; n += theName[14];
    return n;
  }
  return theName;
}


string Hilbert3DCurve::propertyTable() const
{
  return properties.table();
}


string Hilbert3DCurve::info(ifstream* curveInfo) const
{
  if (is(Invalid)) return string("Not a curve");

  stringstream s;
  s << "Curve name: " << name() << endl; 
  s << "Short form: " << shortName() << endl;

  bool found = false;
  char line[1024];
  
  if (curveInfo)
  {
    curveInfo->clear();
    curveInfo->seekg(0);
    string lookFor = string("@") + shortName() + string("@");
    while (!found && curveInfo->good()) 
    { 
      curveInfo->getline(line, 1024);
      found = (string(line).substr(0,lookFor.length()) == lookFor);
    }
  }
  
  if (found)
  {
    while (curveInfo->good())
    {
      curveInfo->getline(line, 1024);
      if (line[0] == '@') break;
      s << line << endl;
    }
  }
  
  s << properties.report();
  
  s << "Definition: " << description() << endl;

  vector<OPt> v; generateTraversal(v, 2);
  s << "4x4x4 grid traversal order:" << endl;
  for (int i = 0; i < 64; ++i) { s << convert(v[i], 2); if (i % 8 == 7) s << endl; }
  if (is(VertexGated) && !is(Symmetric) && theName[1] <= 'o' && theName[2] == theName[3])
    // asymmetric vertex-gated curve with a symmetric base pattern:
    s << "Note: there are other curves that traverse the 4x4x4 grid in the same order." << endl;
  else
    s << "This traversal order uniquely defines this octant-wise self-similar continuous traversal." << endl;
  
  s << endl;
  
  return s.str();
}
