/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

#ifndef HILBERT3DSUITE_TRAVERSAL
#define HILBERT3DSUITE_TRAVERSAL

#include <vector>
using std::vector;
#include <bitset>
using std::bitset;
#include "geometry.h"
#include "properties.h" 

class RecursiveTraversal;
class LimitedTraversal;

class RecursiveTraversal
{
  protected:
    OPt basePattern[8];
    int invBasePattern[2][2][2];
    Transformation transformationPattern[8];

  public:
    RecursiveTraversal();
    RecursiveTraversal(istream& s);
    string description(const Transformation& t = Transformation()) const;
    void generateTraversal(
      vector<OPt>& v, 
      int depth = 3, 
      const Transformation& t = Transformation(), 
      const OPt offset = OPt(0,0,0)
    ) const;
/*
    int index(int depth, NPt p) const;
    NPt point(int depth, int i) const;
*/
};

// how to operate on it using a transformation

class TransformationCandidates
{
  private:
    GeometricTransformation geometry[2];
    int count[2];
  public:
    TransformationCandidates();
    int nrCandidates() const;
    int nrCandidates(const int direction) const;
    Transformation candidate(const int direction) const;
  friend LimitedTraversal;
};

class LimitedTraversal
{
  private:
    typedef vector<OPt> Approximant;
    vector<Approximant> a;
    TransformationCandidates candidates[8];
    Properties properties;
    string prefix;
    void calculateCanonicalForm(const Transformation& embedding, const Transformation& symmetry, vector<Approximant>& canonical, OPt localGate[2][5]);
    void analyseProperties();
  public:
    LimitedTraversal();
    LimitedTraversal(istream& s);
    bool match(const RecursiveTraversal& r, Transformation& t) const;
    bool requires(const Property p) const;
    bool forbids(const Property p) const;
    string name() const;
    const TransformationCandidates& transformation(const int i) const;
    
  friend ostream& operator<<(ostream& o, const LimitedTraversal& t);
};

ostream& operator<<(ostream& o, const LimitedTraversal& t);

#endif

